#!/usr/bin/python
# openTubes.py version 0.1 (Python 2.7)
# Reads the input file openTubes.inp and writes a blockMeshDict file for use in OpenFOAM.
# R.G. Bankras (radko@bankras.net) Copyright 2015

# Inspired by 'Automating blockMesh pipe geometries in OpenFOAM' by Kjetil Moe.
# http://www.kjetilbm.net/blogg/automating-blockmesh-pipe-geometries-in-openfoam/
# http://versati.no/mesher/

# Description of input format 'openTubes.inp'
#       line 1: (round/square),(nr cells square core),(nr cells shell),(vector first tube <u, v, w>),(diameter tube at start)
#       line N: (round/square),(vector in plane of bend <u, v, w>),(diameter bend or 0 for straight),(angle bend or length straight),(diameter tube at end),(nr cells in length)

import math
import sys

def readInputFile():
    "This function reads input data from openTubes.inp and parses the data to lists."
    listSettingsOpenTubes = []
    listTypeOpenTubes = []
    listOpenTubes = []
    try:
        # Open file and read all lines as list of strings
        fileOpenTubes = open('openTubes.inp', 'r')
        with fileOpenTubes:
            stringsOpenTubes = fileOpenTubes.readlines()
        fileOpenTubes.close()
    except:
        print '!! readInputFile(): unable to read file openTubes.inp'
    try:
        # Remove end-of-line characters from strings
        stringsOpenTubes = [i.rstrip('\n') for i in stringsOpenTubes]
        # Split comma-separated strings to list of lists
        listOpenTubes = [i.split(',') for i in stringsOpenTubes]
        # Extract types of tube sections (R = round, S = square)      
        listTypeOpenTubes = [s[0] for s in listOpenTubes]
        # Convert values to type float
        listOpenTubes[:] = [map(float, i) for i in (s[1:] for s in listOpenTubes)]
        # Get initialization settings from first line (
        listSettingsOpenTubes = listOpenTubes.pop(0)
        intCountTubeSections = len(listOpenTubes)
        print '-- Input data read from file openTubes.inp'
    except:
        print '!! readInputFile(): unable to parse data from openTubes.inp'
    return listSettingsOpenTubes, listTypeOpenTubes, listOpenTubes
    
def writeHeader():
    "This function writes the header of the blockMeshDict file."
    try:
        fileBlockMeshDict = open('blockMeshDict', 'w')
        fileBlockMeshDict.write('/*--------------------------------*- C++ -*--------------------------------*\\ \n')
        fileBlockMeshDict.write('|   \\\\   O     T otally      | openTubes: The Open Source tube mesh creator | \n')
        fileBlockMeshDict.write('|    ))  \\\\    U sefull      | Version:   0.1                               | \n')
        fileBlockMeshDict.write('|   ((    ))   B ended       | Web:       www.bankras.org                   | \n')
        fileBlockMeshDict.write('|    \\\\  ((    E xtentions   |                                              | \n')
        fileBlockMeshDict.write('|     \\\\  []    S quare core |                                              | \n')
        fileBlockMeshDict.write('\\*-------------------------------------------------------------------------*/  \n')
        fileBlockMeshDict.write('\n')
        fileBlockMeshDict.write('FoamFile\n')
        fileBlockMeshDict.write('{\n')
        fileBlockMeshDict.write('    version     2.0;\n')
        fileBlockMeshDict.write('    format      ascii;\n')
        fileBlockMeshDict.write('    class       dictionary;\n')
        fileBlockMeshDict.write('    object      blockMeshDict;\n')
        fileBlockMeshDict.write('}\n')
        fileBlockMeshDict.write('\n')
        fileBlockMeshDict.write('convertToMeters 1e-3;\n')
        fileBlockMeshDict.write('\n')
        fileBlockMeshDict.close()
        print '-- Write header to file blockMeshDict'
    except:
        print '!! Unable to open file blockMeshDict'
    return



def writeVertices(listVertices):
    "This function writes the vertices to the blockMeshDict file."
    try:
        fileBlockMeshDict = open('blockMeshDict', 'a')
        fileBlockMeshDict.write('vertices\n')
        fileBlockMeshDict.write('(\n')
        for i in xrange(len(listVertices)):
            fileBlockMeshDict.write('    (' + ' '.join(str(x) for x in listVertices[i]) + ') // point %i\n' % i)
        fileBlockMeshDict.write(');\n')
        fileBlockMeshDict.write('\n')
        fileBlockMeshDict.close()
        print '-- Write vertices to file blockMeshDict'
    except:
        print '!! writeVertices(): unable to open file blockMeshDict'
    return

def writeBlocks(intCountTubeSections, intNumCellsSquareCore, intNumCellsShell, listNumCellsLength):
    "This function writes the blocks to the blockMeshDict file."
    try:
        fileBlockMeshDict = open('blockMeshDict', 'a')
        fileBlockMeshDict.write('blocks\n')
        fileBlockMeshDict.write('(\n')
        listBlocks = ((1, 0, 3, 2, 9, 8, 11, 10), (5, 4, 0, 1, 13, 12, 8, 9), (1, 2, 6, 5, 9, 10, 14, 13), (2, 3, 7, 6, 10, 11, 15, 14), (3, 0, 4, 7, 11, 8, 12, 15))
        for s in xrange(intCountTubeSections):
            for n in xrange(len(listBlocks)):
                fileBlockMeshDict.write('    hex (' + ' '.join(str(x) for x in [i + (s * 8) for i in listBlocks[n]]) + \
                    ') (%i %i %i) simpleGrading (1 1 1) // block %i\n' % (intNumCellsSquareCore, intNumCellsSquareCore, listNumCellsLength[s], 5 * s + n))
        fileBlockMeshDict.write(');\n')
        fileBlockMeshDict.write('\n')
        fileBlockMeshDict.close()
        print '-- Write blocks to file blockMeshDict'
    except:
        print '!! writeBlocks(): unable to open file blockMeshDict'
    return    

def writeEdges(intCountTubeSections, listTypeSections, listPerimeterEdges, listBendEdges):
    "This function writes the edges to the blockMeshDict file."
    try:
        p = 0
        fileBlockMeshDict = open('blockMeshDict', 'a')
        fileBlockMeshDict.write('edges\n')
        fileBlockMeshDict.write('(\n')
        listEdges = ((4, 5), (5, 6), (6, 7), (7, 4))
        for s in xrange(intCountTubeSections + 1):
            # Write the perimeter edges in case the tube section is round
            if (listTypeSections[s] == 'R'):            
                for n in xrange(len(listEdges)):
                    fileBlockMeshDict.write('    arc ' + ' '.join(str(x) for x in [i + (s * 8) for i in listEdges[n]]) + \
                        ' (' + ' '.join(str(x) for x in listPerimeterEdges[s * len(listEdges) + n]) + ') // edge %i\n' % p)
                    p += 1
        listEdges = ((4, 12), (5, 13), (6, 14), (7, 15))
        for s in xrange(intCountTubeSections):
            for n in xrange(len(listEdges)):
                # Only write arc for bend tube sections
                if (len(listBendEdges[s * len(listEdges) + n]) > 1):
                    fileBlockMeshDict.write('    arc ' + ' '.join(str(x) for x in [i + (s * 8) for i in listEdges[n]]) + \
                        ' (' + ' '.join(str(x) for x in listBendEdges[s * len(listEdges) + n]) + ') // edge %i\n' % p)
                    p += 1                
        fileBlockMeshDict.write(');\n')
        fileBlockMeshDict.write('\n')
        fileBlockMeshDict.close()
        print '-- Write edges to file blockMeshDict'
    except:
        print '!! writeEdges(): unable to open file blockMeshDict'
    return    

def writeBoundary(intCountTubeSections):
    "This function writes the boundary to the blockMeshDict file."
    try:
        p = 0
        fileBlockMeshDict = open('blockMeshDict', 'a')
        fileBlockMeshDict.write('boundary\n(\n')
        fileBlockMeshDict.write('    inlet\n    {\n        type patch;\n        faces\n        (\n')
        listEndPatches = ((0, 3, 2, 1), (0, 4, 7, 3), (4, 0, 1, 5), (1, 2, 6, 5), (3, 7, 6, 2))        
        for n in xrange(len(listEndPatches)):
            fileBlockMeshDict.write('            (' + ' '.join(str(x) for x in [i for i in listEndPatches[n]]) + ') // patch %i\n' % p)
            p += 1
        fileBlockMeshDict.write('        );\n    }\n')
        fileBlockMeshDict.write('    walls\n    {\n        type wall;\n        faces\n        (\n')
        listWallPatches = ((5, 4, 12, 13), (5, 13, 14, 6), (6, 14, 15, 7), (7, 15, 12, 4))
        for s in xrange(intCountTubeSections):
            for n in xrange(len(listWallPatches)):
                fileBlockMeshDict.write('            (' + ' '.join(str(x) for x in [i + (s * 8) for i in listWallPatches[n]]) + ') // patch %i\n' % p)
                p += 1
        fileBlockMeshDict.write('        );\n    }\n')
        fileBlockMeshDict.write('    outlet\n    {\n        type patch;\n        faces\n        (\n')
        s = intCountTubeSections
        for n in xrange(len(listEndPatches)):
            fileBlockMeshDict.write('            (' + ' '.join(str(x) for x in [i + (s * 8) for i in listEndPatches[n]]) + ') // patch %i\n' % p)
            p += 1
        fileBlockMeshDict.write('        );\n    }\n')
        fileBlockMeshDict.write(');\n')
        fileBlockMeshDict.write('\n')
        fileBlockMeshDict.close()
        print '-- Write boundary to file blockMeshDict'
    except:
        print '!! writeBoundary(): unable to open file blockMeshDict'
    return   

def writeMergePatchPairs():
    "This function writes the mergePatchPairs to the blockMeshDict file."
    # Face matching does not require internal patches to be defined or pairs to be listed.
    try:
        fileBlockMeshDict = open('blockMeshDict', 'a')
        fileBlockMeshDict.write('mergePatchPairs\n')
        fileBlockMeshDict.write('(\n')
        fileBlockMeshDict.write(');\n')
        fileBlockMeshDict.write('\n')
        fileBlockMeshDict.close()
        print '-- Write mergePatchPairs to file blockMeshDict'
    except:
        print '!! writeMergePatchPairs(): unable to open file blockMeshDict'
    return  

def rotatePointOverAxis(point, angle, axis):
    "This function returns the result of rotating point (x, y, z) over an axis through point (a, b, c) with direction <u, v, w>. \
    The direction of rotation follows the right-hand rule."
    x, y, z = point
    theta = math.radians(angle)
    a, b, c, u, v, w = axis
    l = u * u + v * v + w * w
    result = []
    result.append(((a * (v * v + w * w) - u * (b * v + c * w - u * x - v * y - w * z)) * (1 - math.cos(theta)) + \
        l * x * math.cos(theta) + math.sqrt(l) * (-c * v + b * w - w * y + v * z) * math.sin(theta)) / l)
    result.append(((b * (u * u + w * w) - v * (a * u + c * w - u * x - v * y - w * z)) * (1 - math.cos(theta)) + \
        l * y * math.cos(theta) + math.sqrt(l) * (c * u - a * w + w * x - u * z) * math.sin(theta)) / l)
    result.append(((c * (u * u + v * v) - w * (a * u + b * v - u * x - v * y - w * z)) * (1 - math.cos(theta)) + \
        l * z * math.cos(theta) + math.sqrt(l) * (-b * u + a * v - v * x + u * y) * math.sin(theta)) / l)
    return result

def rotateVectorOverAxis(vector, angle, axis):
    "This function returns the result of rotating vector/point <r, s, t> over an axis through point (0, 0, 0) with direction <u, v, w>. \
    The direction of rotation follows the right-hand rule."
    r, s, t = vector
    theta = math.radians(angle)
    u, v, w = unitVector(axis[3:6])
    result = []
    result.append((math.cos(theta) + u * u * (1 - math.cos(theta))) * r + \
        (u * v * (1 - math.cos(theta)) - w * math.sin(theta)) * s + \
        (u * w * (1 - math.cos(theta)) + v * math.sin(theta)) * t)
    result.append((u * v * (1 - math.cos(theta)) + w * math.sin(theta)) * r + \
        (math.cos(theta) + v * v * (1 - math.cos(theta))) * s + \
        (v * w * (1 - math.cos(theta)) - u * math.sin(theta)) * t)
    result.append((u * w * (1 - math.cos(theta)) - v * math.sin(theta)) * r + \
        (v * w * (1 - math.cos(theta)) + u * math.sin(theta)) * s + \
        (math.cos(theta) + w * w * (1 - math.cos(theta))) * t)
    return result

def unitVector(vector):
    "This function returns the unit vector of vector <u, v, w>."
    u, v, w = vector
    magnitude = math.sqrt(u * u + v * v + w * w)
    result = []
    result.append(u / magnitude)
    result.append(v / magnitude)
    result.append(w / magnitude)
    return result
    
def transformPointByVector(point, vector, length):
    "This function returns the result of transforming point (x, y, z) by vector <r, s, t> over length l."
    x, y, z = point
    r, s, t = unitVector(vector)
    l = length
    result = []
    result.append(x + l * r)
    result.append(y + l * s)
    result.append(z + l * t)
    return result
   
def calcCrossProduct(vector1, vector2):
    "This function results the cross product (pseudonormal) of vector1 <r, s, t> and vector2 <u, v, w>."
    r, s, t = vector1
    u, v, w = vector2
    result = []
    result.append(s * w - t * v)
    result.append(t * u - r * w)
    result.append(r * v - s * u)
    return result

def calcPlaneNormalVectors(vector1, vector2):
    "This function returns the normal of the plane defined by vector1 and vector2."
    theta = calcAngleVectors(vector1, vector2)
    pseudoNormal = calcCrossProduct(vector1, vector2)
    pseudoNormalAxis = [0.0, 0.0, 0.0]
    pseudoNormalAxis.extend(pseudoNormal)
    vector3 = rotateVectorOverAxis(vector1, 1.0, pseudoNormalAxis)
    theta2 = calcAngleVectors(vector2, vector3)
    result = []
    if (theta2 < theta):
        result = pseudoNormal
    else:
        result[:] = [-i for i in pseudoNormal]
    return unitVector(result)
    
def calcBendCenter(point, vector1, vector2, diameter):
    "This function returns the center point coordinates of the bend with diameter d, which begins at point (x, y, z) in the direction of \
        vector1 <r, s, t>. Vector2 <k, l, m> completes the definition of the plane of the bend, which has axis of rotation normal n. The \
        direction of the center point is defined by vector <u, v, w>, which is found as normal of the axis of rotation and vector1."
    x, y, z = point
    r, s, t = vector1
    d = diameter
    n = calcPlaneNormalVectors(vector1, vector2)
    u, v, w = unitVector(calcPlaneNormalVectors(n, vector1))
    result = []
    result.append(x + d / 2 * u)
    result.append(y + d / 2 * v)
    result.append(z + d / 2 * w)
    return result

def scalePointToCenter(point1, point2, factor):
    "This function scales point1 with respect to point2 by a factor f."
    x, y, z = point1
    a, b, c = point2
    f = factor
    result = []
    result.append(a + f * (x - a))
    result.append(b + f * (y - b))
    result.append(c + f * (z - c))
    return result

def calcAngleVectors(vector1, vector2):
    "This function calculates the angle between vector1 <r, s, t> and vector2 <u, v, w>."
    r, s, t = unitVector(vector1)
    u, v, w = unitVector(vector2)
    result = math.degrees(math.acos(r * u + s * v + t * w))
    if (result < 0):
        result += 180
    if (result > 180):
        result -= 180
    return result


# Initialize lists
listSettings = []
listTubeSections = []
listVectors = [] 
listBeginVertices = []
listPerimeterEdges = []
listBendEdges = []
listVertices = []
listEndVertices = []
listAxisRotation = []
listNumCellsLength = []

# Read the input data
listSettings, listTypeSections, listTubeSections = readInputFile()

# Parse settings to variables
intNumCellsSquareCore = int(listSettings[0])
intNumCellsShell = int(listSettings[1])
listVectors.append(listSettings[2:5])
floatDiameterBegin = listSettings[5]

# Count number of tube sections
intCountTubeSections = len(listTubeSections)

# Calculate the offsets from center for the square core and shell
floatRadiusBegin = floatDiameterBegin / 2.0
floatOffsetSquareCoreBegin = floatRadiusBegin / 2.0
floatOffsetShellBegin = floatRadiusBegin * math.cos(math.radians(45.0))

# Set the center coordinates of the start of the first tube section
listCenterPointBegin = [0, 0, 0]

# List the vertices of the inlet
listBeginVertices.append([floatOffsetSquareCoreBegin, 0.0, floatOffsetSquareCoreBegin])
listBeginVertices.append([-floatOffsetSquareCoreBegin, 0.0, floatOffsetSquareCoreBegin])
listBeginVertices.append([-floatOffsetSquareCoreBegin, 0.0, -floatOffsetSquareCoreBegin])
listBeginVertices.append([floatOffsetSquareCoreBegin, 0.0, -floatOffsetSquareCoreBegin])
listBeginVertices.append([floatOffsetShellBegin, 0.0, floatOffsetShellBegin])
listBeginVertices.append([-floatOffsetShellBegin, 0.0, floatOffsetShellBegin])
listBeginVertices.append([-floatOffsetShellBegin, 0.0, -floatOffsetShellBegin])
listBeginVertices.append([floatOffsetShellBegin, 0.0, -floatOffsetShellBegin])

# List the vertices of the inlet for a round circumference
listPerimeterEdges.append([0.0, 0.0, floatRadiusBegin])
listPerimeterEdges.append([-floatRadiusBegin, 0.0, 0.0])
listPerimeterEdges.append([0.0, 0.0, -floatRadiusBegin])
listPerimeterEdges.append([floatRadiusBegin, 0.0, 0.0])

# Determine the angle and axis of rotation for the initial vertices with respect to <0, 1, 0>
if (calcCrossProduct([0.0, 1.0, 0.0], listVectors[0]) != [0.0, 0.0, 0.0]):
    floatTheta = calcAngleVectors([0.0, 1.0, 0.0], listVectors[0])
    listAxisRotation.extend([0.0, 0.0, 0.0])
    listAxisRotation.extend(calcPlaneNormalVectors([0.0, 1.0, 0.0],  listVectors[0]))
    for i in xrange(len(listBeginVertices)):
        point = rotatePointOverAxis(listBeginVertices[i], floatTheta, listAxisRotation)
        listBeginVertices[i] = point
    for i in xrange(len(listPerimeterEdges)):
        point = rotatePointOverAxis(listPerimeterEdges[i], floatTheta, listAxisRotation)
        listPerimeterEdges[i] = point

# Keep track of all new vertices for writing to the blockMeshDict file
listVertices.extend(listBeginVertices)

# Determine the other vertices of the tube sections
for i in xrange(intCountTubeSections):
    print '-- Start tube section: %i' % i
    
    # Reset lists
    del listEndVertices[:]
    del listAxisRotation[:]
    
    # Straight or bend section
    if (listTubeSections[i][3] == 0):
        boolStraightSection = True
    else:
        boolStraightSection = False
    
    # How many cells in this tube section
    listNumCellsLength.append(listTubeSections[i][6])
    
    # Calculate the angle between the start vector listVectors[i] and the bend vector listTubeSections[i][0:3]
    if not(boolStraightSection):
        angleVectors = calcAngleVectors(listVectors[i], listTubeSections[i][0:3])
        if (angleVectors == 0):
            print 'XX Vector for direction of bend is equal to vector of start'
            sys.exit(1)

    # Determine the axis of rotation for the bend
    if not(boolStraightSection):
        listAxisRotation.extend(calcBendCenter(listCenterPointBegin, listVectors[i], listTubeSections[i][0:3], listTubeSections[i][3]))
        listAxisRotation.extend(calcPlaneNormalVectors(listVectors[i], listTubeSections[i][0:3]))
        print '?? listAxisRotation: ' + str(listAxisRotation)

    # Determine the angle of the bend or length of straight section
    if boolStraightSection:
        length = listTubeSections[i][4]
        print '?? length: %.2f' % length
    else:
        floatTheta = listTubeSections[i][4]
        print '?? floatTheta: %.2f' % floatTheta

    # Determine the center points in the middle and at the end of the tube section
    if boolStraightSection:
        centerPointMid = transformPointByVector(listCenterPointBegin, listVectors[i], length / 2.0)
        centerPointEnd = transformPointByVector(listCenterPointBegin, listVectors[i], length)
    else:
        centerPointMid = rotatePointOverAxis(listCenterPointBegin, floatTheta / 2.0, listAxisRotation)
        centerPointEnd = rotatePointOverAxis(listCenterPointBegin, floatTheta, listAxisRotation)
    print '?? listCenterPointBegin: ' + str(listCenterPointBegin)
    print '?? centerPointMid: ' + str(centerPointMid)
    print '?? centerPointEnd: ' + str(centerPointEnd)

    # Get the diameter of the end of the tube section
    diameterEnd = listTubeSections[i][5]

    # Determine the vertices of the end of the tube section
    for n in xrange(len(listBeginVertices)):
        if boolStraightSection:
            listEndVertices.append(scalePointToCenter(transformPointByVector(listBeginVertices[n], listVectors[i], length), \
                centerPointEnd, diameterEnd / floatDiameterBegin))
        else:
            listEndVertices.append(scalePointToCenter(rotatePointOverAxis(listBeginVertices[n], floatTheta, listAxisRotation), \
                centerPointEnd, diameterEnd / floatDiameterBegin))
    for n in xrange(4):
        if boolStraightSection:
            listPerimeterEdges.append(scalePointToCenter(transformPointByVector(listPerimeterEdges[len(listPerimeterEdges) - 4], \
                listVectors[i], length), centerPointEnd, diameterEnd / floatDiameterBegin))
        else:
            listPerimeterEdges.append(scalePointToCenter(rotatePointOverAxis(listPerimeterEdges[len(listPerimeterEdges) - 4], \
                floatTheta, listAxisRotation), centerPointEnd, diameterEnd / floatDiameterBegin))
    
    # For bend tube sections rotate the 4 points on the perimeter by floatTheta/2 and append to listBendEdges
    for n in xrange(4):
        if boolStraightSection:
            listBendEdges.append([0])
        else:
            listBendEdges.append(scalePointToCenter(rotatePointOverAxis(listBeginVertices[len(listBeginVertices) - 4 + n], \
                floatTheta / 2.0, listAxisRotation), centerPointMid, 0.5 + (diameterEnd / floatDiameterBegin) / 2.0))

    # Determine the vector for the next tube section
    if boolStraightSection:
        listVectors.append(listVectors[i])
    else:
        listVectors.append(rotateVectorOverAxis(listVectors[i], floatTheta, listAxisRotation))

    # Keep track of all new vertices for writing to the blockMeshDict file
    listVertices.extend(listEndVertices)

    # Use the vertices at the end of the tube as start of the next section
    del listBeginVertices[:]    
    listBeginVertices.extend(listEndVertices)
    
    # Use the center point at the end of the tube for the next section
    listCenterPointBegin = centerPointEnd

    # Use the diameter at the end of the tube for the next section
    floatDiameterBegin = diameterEnd

# Write the blockMeshDict file
writeHeader()
writeVertices(listVertices)
writeBlocks(intCountTubeSections, intNumCellsSquareCore, intNumCellsShell, listNumCellsLength)
writeEdges(intCountTubeSections, listTypeSections, listPerimeterEdges, listBendEdges)
writeBoundary(intCountTubeSections)
writeMergePatchPairs()
print '-- Finished writing file blockMeshDict'

